/*
 * Brancher.cpp
 *
 *  Created on: 30.10.2020
 *      Author: bzckaise
 */

#include "Brancher.h"

#include <math.h>


namespace CleanMaple {

Brancher::~Brancher() {
	// TODO Auto-generated destructor stub
}


void Brancher::antiexploration_check(Minisat::Var& v, uint32_t& age)
{
    double decay = pow(0.95, age);
    activity_CHB[v] *= decay;
    if (order_heap_CHB.inHeap(v))
        order_heap_CHB.increase(v);
    canceleds[v] = conflicts;
    v = order_heap_CHB[0];
    age = conflicts - canceleds[v];
}

Minisat::Lit Brancher::pickBranchLit()
{
	Minisat::Var next = var_Undef;
	// FIXME: This is not the right place to pick the correct heap.
    Minisat::Heap<VarOrderLt>& order_heap = DISTANCE ? order_heap_distance : ((!VSIDS)? order_heap_CHB:order_heap_VSIDS);

	{
		// FIXME: Using directive for the Macro l_Undef
		using namespace Minisat;
		// Activity based decision:
		while (next == var_Undef || assignments.value(next) != l_Undef
				|| !decision(next))
			if (order_heap.empty()) {
				return lit_Undef;
			} else {
				if (!VSIDS) {
					Minisat::Var v = order_heap_CHB[0];
					uint32_t age = conflicts - canceleds[v];
					while (age > 0) {
						antiexploration_check(v, age);
					}
				}
				next = order_heap.removeMin();
			}
	}

    return Minisat::mkLit(next, assignments.polarity(next));
}

void Brancher::adjust_after_canceling(const Minisat::Var& x)
{
	 if (!VSIDS){
		uint32_t age = conflicts - pickeds[x];
		if (age > 0){
			double adjusted_reward = ((double) (conflicteds[x] + almost_conflicteds[x])) / ((double) age);
			double old_activity = activity_CHB[x];
			activity_CHB[x] = step_size * adjusted_reward + ((1 - step_size) * old_activity);
			if (order_heap_CHB.inHeap(x)){
				if (activity_CHB[x] > old_activity)
					order_heap_CHB.decrease(x);
				else
					order_heap_CHB.increase(x);
			}
		}
		canceleds[x] = conflicts;
	}

	insertVarOrder(x);
}

void Brancher::rebuildOrderHeap()
{
	Minisat::vec<Minisat::Var> vs;
	for (Minisat::Var v = 0; v < activity_VSIDS.size(); v++) {
		//FIXME: Using directive for Macro l_Undef
		using namespace Minisat;
		if (decision(v) && assignments.value(v) == l_Undef)
			vs.push(v);
	}

    order_heap_CHB  .build(vs);
    order_heap_VSIDS.build(vs);
    order_heap_distance.build(vs);
}

void Brancher::analyze_update_1(const Minisat::Var& v) {
// FIXME: I feel like this belongs here. However, currently I have no acces to
// the level function. It is going to be in VariableDatabase soon, however,
// until then the code is in the analyze funtion.
//		analyze_update_tmp.clear();
//        if (VSIDS){
//        	varBumpActivity(v, .5);
//        	analyze_update_tmp.push(v);
//        }else
//            conflicteds[v]++;
}

void Brancher::analyze_update_2(const int& out_btlevel) {
// FIXME: I feel like this belongs here. However, currently I have no acces to
// the level function. It is going to be in VariableDatabase soon, however,
// until then the code is in the analyze funtion.
//	      for (int i = 0; i < analyze_update_tmp.size(); i++){
//	            Minisat::Var v = analyze_update_tmp[i];
//	            if (level(v) >= out_btlevel - 1)
//	                varBumpActivity(v, 1);
//	        }
//	      analyze_update_tmp.clear();
}

void Brancher::uncheckedEnque_adjustment(Minisat::Var& x)
{
    if (!VSIDS){
        pickeds[x] = conflicts;
        conflicteds[x] = 0;
        almost_conflicteds[x] = 0;
        uint32_t age = conflicts - canceleds[x];
        if (age > 0){
            double decay = pow(0.95, age);
            activity_CHB[x] *= decay;
            if (order_heap_CHB.inHeap(x))
                order_heap_CHB.increase(x);
        }
    }
}

void Brancher::adjust_while_collecting_first_UIP_inner(const Minisat::Var& v, const int& max_level)
{
    activity_distance[v]+=var_iLevel_tmps[v]*level_incs[var_iLevel_tmps[v]-1];

    if(activity_distance[v]>1e100){
        for(int vv=0;vv<activity_distance.size();vv++)
            activity_distance[vv] *= 1e-100;
        var_iLevel_inc*=1e-100;
        for(int j=0; j<max_level; j++) level_incs[j]*=1e-100;
    }
    if (order_heap_distance.inHeap(v))
        order_heap_distance.decrease(v);

}

/*
 * Apparently the order of the variables in involved_lits is really important
 * and resembles something.
 */
void Brancher::adjust_while_collecting_first_UIP_outer(const Minisat::vec<Minisat::Lit>& involved_lits,const int& max_level)
{
    double inc=var_iLevel_inc;
    level_incs.clear();
    for(int i=0;i<max_level;i++){
        level_incs.push(inc);
        inc = inc/my_var_decay;
    }

    for(int i=0;i<involved_lits.size();i++){
        Minisat::Var v =var(involved_lits[i]);
        adjust_while_collecting_first_UIP_inner(v,max_level);
    }
    var_iLevel_inc=level_incs[level_incs.size()-1];
}

void Brancher::Brancher::switch_to_VSIDS() {
    VSIDS = true;
    fflush(stdout);
    pickeds.clear();
    conflicteds.clear();
    almost_conflicteds.clear();
    canceleds.clear();
}

} /* namespace CleanMaple */
