/*
 * Brancher.h
 *
 *  Created on: 30.10.2020
 *      Author: bzckaise
 */

#ifndef CLEANSOLVER_CORESOLVER_BRANCHER_H_
#define CLEANSOLVER_CORESOLVER_BRANCHER_H_

#include "../../Utilities/BooleanTypes/SolverTypes.h"
#include "../../Utilities/Containers/Heap.h"
#include "../../Utilities/Containers/Vec.h"
#include "VariableDatabase.h"

namespace CleanMaple {

class Brancher {
private:

	// FIXME: The following struct VarOrderLT is only disturbing. Get rid of it.
    struct VarOrderLt {
        const Minisat::vec<double>&  activity;
        bool operator () (Minisat::Var x, Minisat::Var y) const { return activity[x] > activity[y]; }
        VarOrderLt(const Minisat::vec<double>&  act) : activity(act) { }
	};

    bool   DISTANCE, VSIDS;
    //FIXME: Rename CHB
    Minisat::Heap<VarOrderLt>    	order_heap_CHB,   // A priority queue of variables ordered with respect to the variable activity.
      	  	  	  	  	  	  	  	order_heap_VSIDS,
									order_heap_distance;
    const VariableDatabase&  		assignments;
    // FIXME: Put the following in some extra class for statistics.
    const uint64_t&					conflicts;
    // FIXME: Do the following really belong here? Problem is, that the Brancher needs
    // write access, so currently they cannot be in the VariableDatabase...
    Minisat::vec<double>         	activity_CHB,     // A heuristic measurement of the activity of a variable.
      	  	  	  	  	  	  	  	activity_VSIDS,
									activity_distance;

    Minisat::vec<uint32_t> 			pickeds;
    Minisat::vec<uint32_t>		 	conflicteds;
    Minisat::vec<uint32_t> 			almost_conflicteds;
    Minisat::vec<uint32_t> 			canceleds;
    double							var_inc;
    Minisat::vec<char>				decisions;         // Declares if a variable is eligible for selection in the decision heuristic.

	// FIXME: Get rid of some of those. Pass the whole container instead.
    uint64_t						dec_vars;
    double							var_decay;
    int       timer;
    Minisat::vec<double> 			var_iLevel_tmps;
    double 							var_iLevel_inc;
    double    						my_var_decay;
    Minisat::vec<int> 				level_incs;
    Minisat::vec<Minisat::Var>		analyze_update_tmp;

    double    step_size;
    double    step_size_dec;
    double    min_step_size;

	inline char	decision         (const Minisat::Var x) const   { return decisions[x]; }




public:
	Brancher(VariableDatabase&  _assignments,
			const uint64_t& _conflicts):
				order_heap_CHB (VarOrderLt(activity_CHB)),
				order_heap_VSIDS (VarOrderLt(activity_VSIDS)),
				order_heap_distance(VarOrderLt(activity_distance)),
				DISTANCE(true),
				VSIDS(true),
				assignments(_assignments),
				conflicts(_conflicts),
				var_inc(1),
				dec_vars(0),
				var_decay(0.80),
				timer(5000),
				var_iLevel_inc(1),
			    my_var_decay(0.6),
			    step_size(0.40),
			    step_size_dec(0.000001),
			    min_step_size(0.06)

		{
			// TODO Auto-generated constructor stub

		};
	virtual ~Brancher();
	Brancher(const Brancher &other) = delete;
	Brancher(Brancher &&other) = delete;
	Brancher& operator=(const Brancher &other) = delete;
	Brancher& operator=(Brancher &&other) = delete;

	void antiexploration_check(Minisat::Var& v, uint32_t& age);
	Minisat::Lit pickBranchLit();

	inline void insertVarOrder(Minisat::Var x) {
	    //    Heap<VarOrderLt>& order_heap = VSIDS ? order_heap_VSIDS : order_heap_CHB;
	    Minisat::Heap<VarOrderLt>& order_heap = DISTANCE ? order_heap_distance : ((!VSIDS)? order_heap_CHB:order_heap_VSIDS);
	    if (!order_heap.inHeap(x) && decision(x)) order_heap.insert(x); }


	inline void varBumpActivity(Minisat::Var v, double mult) {
	    if ( (activity_VSIDS[v] += var_inc * mult) > 1e100 ) {
	        // Re-scale:
	        for (int i = 0; i < activity_VSIDS.size(); i++)
	            activity_VSIDS[i] *= 1e-100;
	        var_inc *= 1e-100; }

	    // Update order_heap with respect to new activity:
	    if (order_heap_VSIDS.inHeap(v)) order_heap_VSIDS.decrease(v); }

	void adjust_after_canceling(const Minisat::Var& x);

	void rebuildOrderHeap();


	void setDecisionVar(Minisat::Var v, bool b) {
	    if (b && !order_heap_CHB.inHeap(v)){
	        order_heap_CHB.insert(v);
	        order_heap_VSIDS.insert(v);
	        order_heap_distance.insert(v);}
	    if      ( b && !decision(v)) dec_vars++;
	    else if (!b &&  decision(v)) dec_vars--;

	    // FIXME: I think it is a bad idea to let the brancher be able to choose the
	    // variables it may choose by itself... I am not sure, whether this is
	    // good...
	    decisions[v] = b;
	}

	void uncheckedEnque_adjustment(Minisat::Var& x);

	void adjust_while_collecting_first_UIP_inner(const Minisat::Var &v, const int& max_level);
	void adjust_while_collecting_first_UIP_outer(
			const Minisat::vec<Minisat::Lit> &involved_lits,
			const int &max_level);


	// FIXME: The following two are not yet implemented due to the fact, that
	// the level function is not yet included in to the VariableDatabase.
	void analyze_update_1(const Minisat::Var& v);
	void analyze_update_2(const int& out_btlevel);



	inline void varDecayActivity() {
	    var_inc *= (1 / var_decay); }

	inline void newVar() {
	    activity_CHB  .push(0);
	    activity_VSIDS.push(0);
	    activity_distance.push(0);
	    canceleds.push(0);
	    decisions.push();
	    pickeds.push(0);
	    conflicteds.push(0);
	    almost_conflicteds.push(0);
	    var_iLevel_tmps.push(0);
	}

	void switch_to_VSIDS();

	inline const uint64_t&    get_dec_vars() const  { return dec_vars; }


	inline uint32_t&    conflicted			(const Minisat::Var x)   { return conflicteds[x]; }
	inline const uint32_t&    conflicted			(const Minisat::Var x)  const { return conflicteds[x]; }

	inline uint32_t&    almost_conflicted	(const Minisat::Var x)   { return almost_conflicteds[x]; }
	inline const uint32_t&    almost_conflicted	(const Minisat::Var x)  const { return almost_conflicteds[x]; }

	// FIXME: Get rid of those. Pass the whole container instead.
	inline double&    var_iLevel_tmp	(const Minisat::Var x)   { return var_iLevel_tmps[x]; }
	inline const double&    var_iLevel_tmp	(const Minisat::Var x)  const { return var_iLevel_tmps[x]; }

	inline void adjust_timer_and_var_decay()
	{
        if (--timer == 0 && var_decay < 0.95) timer = 5000, var_decay += 0.01;

	}

	inline void adjust_stepsize() {
        if (step_size > min_step_size) step_size -= step_size_dec;
	}

	inline bool getDistance() const {
		return DISTANCE;
	}

	inline void setDistance(bool distance) {
		DISTANCE = distance;
	}

	inline bool getVsids() const {
		return VSIDS;
	}

	inline void setVsids(bool vsids) {
		VSIDS = vsids;
	}

};



} /* namespace CleanMaple */

#endif /* CLEANSOLVER_CORESOLVER_BRANCHER_H_ */
