/*
 * ClauseDatabase.cpp
 *
 *  Created on: Nov 12, 2020
 *      Author: kokoro
 */

#include "ClauseDatabase.h"


#include "../../Utilities/Algorithms/Sort.h"
#include "../../Utilities/Containers/Vec.h"

namespace CleanMaple {

ClauseDatabase::~ClauseDatabase() {
	// TODO Auto-generated destructor stub
	// TODO: Can I put the definition into the header? Yes, right?
}

/*
 * TODO: Get rid of this struct.
 */
struct reduceDB_lt {
    Minisat::ClauseAllocator& ca;
    reduceDB_lt(Minisat::ClauseAllocator& ca_) : ca(ca_) {}
    bool operator () (Minisat::CRef x, Minisat::CRef y) const { return ca[x].activity() < ca[y].activity(); }
};

void ClauseDatabase::reduceDB() {
	int i, j;
	sort(learnts_local, reduceDB_lt(ca));

	    int limit = learnts_local.size() / 2;
	    for (i = j = 0; i < learnts_local.size(); i++){
	        Minisat::Clause& c = ca[learnts_local[i]];
	        if (c.mark() == LOCAL)
	            if (c.removable() && !locked(c) && i < limit)
	            	remove_clause_from_database(learnts_local[i]);
	            else{
	                if (!c.removable()) limit++;
	                c.removable(true);
	                learnts_local[j++] = learnts_local[i]; }
	    }
	    learnts_local.shrink(i - j);
}

void ClauseDatabase::reduceDB_Tier2(uint64_t& current_number_of_conflicts) {
    int i, j;
    for (i = j = 0; i < learnts_tier2.size(); i++){
        Minisat::Clause& c = ca[learnts_tier2[i]];
        if (c.mark() == TIER2)
            if (!locked(c) && c.touched() + 30000 < current_number_of_conflicts){
                learnts_local.push(learnts_tier2[i]);
                c.mark(LOCAL);
                c.activity() = 0;
                claBumpActivity(c);
            }else
                learnts_tier2[j++] = learnts_tier2[i];
    }
    learnts_tier2.shrink(i - j);
}

void ClauseDatabase::design_fixing_unlock_if_locked(Minisat::Clause& c) {
	// Original comment:
	// 		"NOTE: Don't leave pointers to free'd memory!"
	// TODO: Rewrite the following if (...) {...}
	// TODO: Find out why it does happen that locked clauses are removed.
	// Is this even a bug? Make sure this does not happen.
	// TODO: Write two versions of this removeClause(), one in which the
	// following can happen, and one where it can not, such that we can not
	// run into surprises, or at least notice them.
	if (locked(c)) {
		unlock(c);
	}
}



void ClauseDatabase::reallocate_all_watches(Minisat::ClauseAllocator& new_ca) {
	watches.cleanAll();
	watches_bin.cleanAll();
	for (int v = 0; v < variable_database.nVars(); v++)
		for (int s = 0; s < 2; s++) {
			Minisat::Lit p = Minisat::mkLit(v, s);
			Minisat::vec<Watcher> &ws = watches[p];
			for (int j = 0; j < ws.size(); j++)
				ca.reloc(ws[j].cref, new_ca);
			Minisat::vec<Watcher> &ws_bin = watches_bin[p];
			for (int j = 0; j < ws_bin.size(); j++)
				ca.reloc(ws_bin[j].cref, new_ca);
		}
}

void ClauseDatabase::reallocate_all_reason_clauses(
		Minisat::ClauseAllocator& new_ca) {
	{
		/* All reasons:
		 * TODO: This is a routine between the implication_graph and the
		 * clause_database... Well, where to put this? Logically, the
		 * implication_graph is not supposed to even notice the relocation. At
		 * the same time, the clause_database should only have reading access to
		 * the implication graph... Well, this is a dilemma.
		 * TODO: Actually, what is needed, is that every class, that uses CRefs,
		 * offers a unique possibility to the allocator to iterate over all CRef's,
		 * where it can reallocate each CRef. Or something similar.
		 */
		for (int i = 0; i < implication_graph.trail.size(); i++) {
			Minisat::Var v = var(implication_graph.trail[i]);
			using namespace Minisat;
			if (implication_graph.reason(v) != CRef_Undef
					&& (ca[implication_graph.reason(v)].reloced()
							|| locked(ca[implication_graph.reason(v)])))
				ca.reloc(implication_graph.reason_data[v].reason, new_ca);
		}
	}
}

void ClauseDatabase::reallocate_all_clauses(Minisat::ClauseAllocator& new_ca) {
	// All learnt:
	for (int i = 0; i < learnts_core.size(); i++)
		ca.reloc(learnts_core[i], new_ca);
	for (int i = 0; i < learnts_tier2.size(); i++)
		ca.reloc(learnts_tier2[i], new_ca);
	for (int i = 0; i < learnts_local.size(); i++)
		ca.reloc(learnts_local[i], new_ca);
	// All original:
	int i, j;
	for (i = j = 0; i < clauses.size(); i++)
		if (ca[clauses[i]].mark() != 1) {
			ca.reloc(clauses[i], new_ca);
			clauses[j++] = clauses[i];
		}
	clauses.shrink(i - j);
}

void ClauseDatabase::reallocate(Minisat::ClauseAllocator& new_ca) {
    // TODO: Instead of call the new_ca by reference as a parameter, this
	// function should start with the following line. However, some stuff from
	// the SimpSolver cannot be put here and also does not belong here, but
	// needs acces to the ClauseAllocator... This is a dilemma.

	// Minisat::ClauseAllocator new_ca(ca.size() - ca.wasted());

	reallocate_all_watches(new_ca);
	reallocate_all_reason_clauses(new_ca);
	reallocate_all_clauses(new_ca);
}

void ClauseDatabase::removeSatisfied(Minisat::vec<Minisat::CRef>& cs) {
    int i, j;
    for (i = j = 0; i < cs.size(); i++){
        Minisat::Clause& c = ca[cs[i]];
        if (variable_database.satisfied(c))
        	remove_clause_from_database(cs[i]);
        else
            cs[j++] = cs[i];
    }
    cs.shrink(i - j);
}

void ClauseDatabase::removeSatisfied() {
    // Remove satisfied clauses:
    removeSatisfied(learnts_core); // Should clean core first.
    safeRemoveSatisfied(learnts_tier2, TIER2);
    safeRemoveSatisfied(learnts_local, LOCAL);
    if (remove_satisfied)        // Can be turned off.
    	removeSatisfied(clauses);
}



void ClauseDatabase::safeRemoveSatisfied(Minisat::vec<Minisat::CRef>& cs,
		unsigned valid_mark) {
	    int i, j;
	    for (i = j = 0; i < cs.size(); i++){
	    	Minisat::Clause& c = ca[cs[i]];
	        if (c.mark() == valid_mark)
	            if (variable_database.satisfied(c))
	            	remove_clause_from_database(cs[i]);
	            else
	                cs[j++] = cs[i];
	    }
	    cs.shrink(i - j);
}

} /* namespace CleanMaple */
