/*
 * ImplicationGraph.h
 *
 *  Created on: Nov 14, 2020
 *      Author: kokoro
 */

#ifndef CLEANSOLVER_CORESOLVER_IMPLICATIONGRAPH_H_
#define CLEANSOLVER_CORESOLVER_IMPLICATIONGRAPH_H_

#include "../../Utilities/BooleanTypes/SolverTypes.h"
#include "../../Utilities/Containers/Vec.h"

namespace CleanMaple {

//////////////////////////////////////
// Helper structures:

// TODO: Get rid of these:
struct ReasonData {
	Minisat::CRef reason;
	int level;
};

// TODO: In particular, get rid of this one.
// TODO: I made this const and pass by reference. Is this a problem?
// TODO: This was static before. I changed this. Is this a problem?
inline ReasonData mkReasonData(const Minisat::CRef& cr, const int& l) {
	ReasonData d = { cr, l };
	return d;
};

////////////////////////////////////////////////

class ImplicationGraph {
public:
	// TODO: 	All of the following need to be private. The only purpose of
	//			their publicity is to have compiling versions while being in the
	// 			middle of the process of factoring all that belongs to this
	//			class out.
	Minisat::vec<Minisat::Lit> trail; // Assignment stack; stores all assigments made in the order they were made.
	Minisat::vec<int> trail_lim; // Separator indices for different decision levels in 'trail'.
	Minisat::vec<ReasonData> reason_data; // Stores reason and level for each variable.
    int                 qhead;            // Head of queue (as index into the trail -- no more explicit propagation queue in MiniSat).


public:
	ImplicationGraph() :
			trail(), trail_lim(), qhead(0) {
	}
	;
	virtual ~ImplicationGraph();
	ImplicationGraph(const ImplicationGraph& other) = delete;
	ImplicationGraph(ImplicationGraph&& other) = delete;
	ImplicationGraph& operator=(const ImplicationGraph& other) = delete;
	ImplicationGraph& operator=(ImplicationGraph&& other) = delete;

	inline void newDecisionLevel() {
		trail_lim.push(trail.size());
	}

	inline int decisionLevel() const {
		return trail_lim.size();
	}

	inline int nAssigns() const {
		return trail.size();
	}

	inline Minisat::CRef reason(const Minisat::Var& x) const {
		return reason_data[x].reason;
	}
	inline void set_reason(const Minisat::Var& x, const Minisat::CRef& reason_clause, const int& level) {
		reason_data[x] = mkReasonData(reason_clause,level);
	}
	// FIXME: Why should this function be ever called?
	inline void set_reason(const Minisat::Var& x, const Minisat::CRef& reason_clause) {
		reason_data[x].reason = reason_clause;
	}

	/*
	 * TODO: Get rid of the following and have a proper construction of the
	 * class.
	 */
	inline void newVar() {
		// TODO: Using directive for the macro.
		using namespace Minisat;
	    reason_data.push(mkReasonData(CRef_Undef,0));
	}

	inline int level(Minisat::Var x) const {
		return reason_data[x].level;
	}

};



} /* namespace CleanMaple */

#endif /* CLEANSOLVER_CORESOLVER_IMPLICATIONGRAPH_H_ */
