/*
 * VariableDatabase.h
 *
 *  Created on: 30.10.2020
 *      Author: bzckaise
 */

#ifndef CLEANSOLVER_CORESOLVER_VARIABLEDATABASE_H_
#define CLEANSOLVER_CORESOLVER_VARIABLEDATABASE_H_

#include "../../Utilities/BooleanTypes/SolverTypes.h"
#include "../../Utilities/Containers/Vec.h"

namespace CleanMaple {

/////////////////////////////////////////
//// Forward declarations:
//class ClauseDatabase {
//	void relocate_reason_data();
//};



class VariableDatabase {

private:
	// FIXME: Currently in the Brancher a lot of stuff is inside, which belong
	// here, but cannot, as the brancher needs write access to these, but is
	// supposed to have only read access to this class.
    Minisat::vec<Minisat::lbool>	assigns;          	// The current assignments.
    Minisat::vec<char>				polarities;         // The preferred polarity of each variable.


public:
	VariableDatabase() :
			assigns(/*TODO: ????*/), polarities(/*TODO: ????*/)
	{
		// do nothing
	}
	virtual ~VariableDatabase()
	{
		// do nothing
	};
	VariableDatabase(const VariableDatabase &other) = delete;
	VariableDatabase(VariableDatabase &&other) = delete;
	VariableDatabase& operator=(const VariableDatabase &other) = delete;
	VariableDatabase& operator=(VariableDatabase &&other) = delete;

	inline void newVar(const bool& sign) {
		{
			// TODO: Using directive only for the macro.
			using namespace Minisat;
			assigns.push(l_Undef);
		}
	    polarities.push(sign);
	}

	/*
	 * TODO: Make all of these getters and setters easier to use. Possibly,
	 * make these return a reference to the underlying object or something like
	 * this.
	 */

	/*
	 * TODO: Make setters and restrict access to these setters to the propagator
	 * and the analyzer, or even less. Possibly have them as non-member friends
	 * or something.
	 */
	inline Minisat::lbool value(const Minisat::Var x) const {
		return assigns[x];
	}
	inline Minisat::lbool value(const Minisat::Lit p) const {
		return assigns[var(p)] ^ sign(p);
	}
	inline void set_value(const Minisat::Var& x, const Minisat::lbool& value) {
		assigns[x] = value;
	}

	/*
	 * This comment is for the following two functions.
	 * TODO: Why is this a vector of char? This is logically not a char, but a
	 * bool (or possibly an lbool).
	 */
	inline char polarity(const Minisat::Var& x) const {
		return polarities[x];
	}
	inline void set_polarity(const Minisat::Var& v,const bool& b) {
		polarities[v] = b;
	}

	/*
	 * TODO: Rewrite the following function.
	 */
	inline int nVars() const {
		return assigns.size();
	}

	inline bool satisfied(const Minisat::Clause& c) const {
	    for (int i = 0; i < c.size(); i++) {
	    	using namespace Minisat;
	        if (value(c[i]) == l_True)
	            return true;
	    }
	    return false; }


	// friend void ClauseDatabase::relocate_reason_data();
};

} /* namespace CleanMaple */

#endif /* CLEANSOLVER_CORESOLVER_VARIABLEDATABASE_H_ */
