/*
 * DratPrinter.h
 *
 *  Created on: 27.10.2020
 *      Author: bzckaise
 */

#ifndef UTILITIES_PRINTUTILITIES_DRATPRINTER_DRATPRINTER_H_
#define UTILITIES_PRINTUTILITIES_DRATPRINTER_DRATPRINTER_H_

#define BIN_DRUP

#include <cstdio>

#include "../../BooleanTypes/SolverTypes.h"


namespace CleanMaple {

class DratPrinter {
public:
	// TODO: 	All of the following need to be private. The only purpose of
	//			their publicity is to have compiling versions while being in the
	// 			middle of the process of factoring all that belongs to this
	//			class out.
    FILE*     drup_file;

public:
    /*
     * TODO: As far as I know this initialization of the drup_file with NULL is
     * currently necessary. Rewrite this somehow.
     */
	DratPrinter() : drup_file(NULL) {};
	virtual ~DratPrinter();
	DratPrinter(const DratPrinter& other) = delete;
	DratPrinter(DratPrinter&& other) = delete;
	DratPrinter& operator=(const DratPrinter& other) = delete;
	DratPrinter& operator=(DratPrinter&& other) = delete;

void removeClause(const Minisat::Clause& c) {
	if (drup_file){
        if (c.mark() != 1){
#ifdef BIN_DRUP
            binDRUP('d', c, drup_file);
#else
            fprintf(drup_file, "d ");
            for (int i = 0; i < c.size(); i++)
                fprintf(drup_file, "%i ", (var(c[i]) + 1) * (-2 * sign(c[i]) + 1));
            fprintf(drup_file, "0\n");
#endif
        }else
            printf("c Bug. I don't expect this to happen.\n");
    }
}
/*
 * TODO: Well, the Macro is entirely unnecessary. Remove it. Make this an option
 * or something similar. Also, why are these static? Rewrite this or provide a
 * reason. Also make these private!
 */
#ifdef BIN_DRUP
    static int buf_len;
    static unsigned char drup_buf[];
    static unsigned char* buf_ptr;

    static inline void byteDRUP(Minisat::Lit l){
        unsigned int u = 2 * (var(l) + 1) + sign(l);
        do{
            *buf_ptr++ = u & 0x7f | 0x80; buf_len++;
            u = u >> 7;
        }while (u);
        *(buf_ptr - 1) &= 0x7f; // End marker of this unsigned number.
    }

    template<class V>
    static inline void binDRUP(unsigned char op, const V& c, FILE* drup_file){
        assert(op == 'a' || op == 'd');
        *buf_ptr++ = op; buf_len++;
        for (int i = 0; i < c.size(); i++) byteDRUP(c[i]);
        *buf_ptr++ = 0; buf_len++;
        if (buf_len > 1048576) binDRUP_flush(drup_file);
    }

    static inline void binDRUP_strengthen(const Minisat::Clause& c, Minisat::Lit l, FILE* drup_file){
        *buf_ptr++ = 'a'; buf_len++;
        for (int i = 0; i < c.size(); i++)
            if (c[i] != l) byteDRUP(c[i]);
        *buf_ptr++ = 0; buf_len++;
        if (buf_len > 1048576) binDRUP_flush(drup_file);
    }

    static inline void binDRUP_flush(FILE* drup_file){
//        fwrite(drup_buf, sizeof(unsigned char), buf_len, drup_file);
        fwrite_unlocked(drup_buf, sizeof(unsigned char), buf_len, drup_file);
        buf_ptr = drup_buf; buf_len = 0;
    }
#endif

};

} /* namespace CleanMaple */




#endif /* UTILITIES_PRINTUTILITIES_DRATPRINTER_DRATPRINTER_H_ */

