#include "LogfilePrinter.h"

void print_initial() {
	initial_time = Minisat::cpuTime();
	printf("c This is MapleLCMDistChronoBT.\n");
	printf(
			"c ============================[ Problem Statistics ]=============================\n");
	printf(
			"c |                                                                             |\n");
}

void printStats(const Minisat::Solver &solver) {
	double cpu_time = Minisat::cpuTime();
	double mem_used = Minisat::memUsedPeak();
	printf("c restarts              : %"PRIu64"\n", solver.starts);
	printf("c conflicts             : %-12"PRIu64"   (%.0f /sec)\n",
			solver.conflicts, solver.conflicts / cpu_time);
	printf(
			"c decisions             : %-12"PRIu64"                  (%.0f /sec)\n",
			solver.decisions,
			solver.decisions / cpu_time);
	printf("c propagations          : %-12"PRIu64"   (%.0f /sec)\n",
			solver.propagator.propagations, solver.propagator.propagations / cpu_time);
	printf("c conflict literals     : %-12"PRIu64"   (%4.2f %% deleted)\n",
			solver.tot_literals,
			(solver.max_literals - solver.tot_literals) * 100
					/ (double) solver.max_literals);
	printf(
			"c backtracks            : %-12"PRIu64"   (NCB %0.f%% , CB %0.f%%)\n",
			solver.non_chrono_backtrack + solver.chrono_backtrack,
			(solver.non_chrono_backtrack * 100)
					/ (double) (solver.non_chrono_backtrack
							+ solver.chrono_backtrack),
			(solver.chrono_backtrack * 100)
					/ (double) (solver.non_chrono_backtrack
							+ solver.chrono_backtrack));
	if (mem_used != 0)
		printf("c Memory used           : %.2f MB\n", mem_used);

	printf("c CPU time              : %g s\n", cpu_time);
}

void print_after_parsing_Dimacs(const Minisat::Solver &S) {
	parsed_time = Minisat::cpuTime();
	if (S.verbosity > 0) {
		printf(
				"c |  Number of variables:  %12d                                         |\n",
				S.variable_database.nVars());
		printf(
				"c |  Number of clauses:    %12d                                         |\n",
				S.clause_database.clauses.size());
	}
	if (S.verbosity > 0)
		printf(
				"c |  Parse time:           %12.2f s                                       |\n",
				parsed_time - initial_time);
}

void print_after_preprocessor(const Minisat::Solver &S) {
	prepro_time = Minisat::cpuTime();
	if (S.verbosity > 0) {
		printf(
				"c |  Simplification time:  %12.2f s                                       |\n",
				prepro_time - parsed_time);
		printf(
				"c |                                                                             |\n");
	}
}

void print_logfile_if_unsat(const Minisat::Solver &S) {
	printf("s UNSATISFIABLE\n");
	printStats(S);
}

void print_logfile_if_unsolved(const Minisat::Solver &S) {
	printf("s UNKNOWN\n");
	printStats(S);
}

void print_logfile_unsat_after_simplification(const Minisat::Solver &S) {
	if (S.verbosity > 0) {
		printf(
				"c ===============================================================================\n");
		printf("c Solved by simplification\n");
		printStats(S);
		printf("\n");
	}
	printf("s UNSATISFIABLE\n");
}

void print_logfile_if_model_is_found(const Minisat::Solver &S) {
	printf("s SATISFIABLE\n");
	printf("v ");
	for (int i = 0; i < S.variable_database.nVars(); i++) {
		// The following using directive is necessary because l_Undef is a
		// Macro. I don't want to change that, yet. FIXME
		using namespace Minisat;
		if (S.model[i] != l_Undef)
			printf("%s%s%d", (i == 0) ? "" : " ",
					(S.model[i] == l_True) ? "" : "-", i + 1);
	}
	printf(" 0\n");
	printStats(S);
}
