#include "decisionaggregate.h"



void 
track_conflict_decision(kissat *solver)
{
    if(solver->decision_conf_data.confs_last_decision==0)
    {
        solver->decision_conf_data.decisions_without_confs++;
        solver->decision_conf_data.succ_no_confs++;
    }
    else
    {
        solver->decision_conf_data.decisions_with_confs++;
        if(solver->decision_conf_data.succ_no_confs)
        {
            solver->decision_conf_data.total_cdp_len +=  solver->decision_conf_data.succ_no_confs;
            solver->decision_conf_data.num_cdp++;
            solver->decision_conf_data.succ_no_confs=0;  
        }
        
        if(solver->decision_conf_data.confs_last_decision>1)
            solver->decision_conf_data.mc_decisions++;
        else
            solver->decision_conf_data.sc_decisions++;   
    }

}

void tag_lbp_vars(kissat *solver)
{
    for (all_stack (unsigned, level,solver->decision_conf_data.lbp))
    {
        frame f = FRAME(level);
        if(VALID_INTERNAL_LITERAL(f.decision)) // must be a valid_internal_literal (skip level=0==>f.decision=-1)
        {
            const unsigned idx = IDX (f.decision);  
            solver->is_crv_decision[idx] = true; 
            solver->min_glue_lbp[idx] = solver->decision_conf_data.min_glue;
        }        
    }
}

bool valid_level(kissat *solver, unsigned level)
{
    return level < solver->level; // "level < solver->level" takes only the reason levels
}

void 
clear_map(kissat* solver, unsigneds st)
{
    for (all_stack (unsigned, i,st))
    {
        if(valid_level(solver,i))       
            binary_map[i]=0;
    }
}

void compute_rest_common(kissat * solver, struct cls * current)
{
    for (all_stack (unsigned, level,solver->decision_conf_data.lbp))
    {
        if(valid_level(solver,level))
            binary_map[level]=1;
    }

    CLEAR_STACK(solver->decision_conf_data.lbp);
    
    for (all_stack (unsigned, level,current->dls))
    {
        if(valid_level(solver,level) && binary_map[level]==1)
            PUSH_STACK(solver->decision_conf_data.lbp,level);
    }

    clear_map(solver, solver->decision_conf_data.lbp);       
}

void compute_first_common(kissat * solver, struct cls * current, struct cls * nxt)
{  
    for (all_stack (unsigned, level,current->dls))
    {
        if(valid_level(solver,level))
            binary_map[level]=1;
    }
    for (all_stack (unsigned, level,nxt->dls))
    {
        if(valid_level(solver,level) && binary_map[level]==1)
            PUSH_STACK(solver->decision_conf_data.lbp,level);
    }
    clear_map(solver, current->dls);
}

void 
compute_lbp(kissat *solver)
{
    struct cls * current=NULL, *nxt=NULL;

    
    if(SIZE_STACK(solver->decision_conf_data.learned_clauses_levels))
        current=POP_STACK(solver->decision_conf_data.learned_clauses_levels);
    if(SIZE_STACK(solver->decision_conf_data.learned_clauses_levels))        
        nxt=POP_STACK(solver->decision_conf_data.learned_clauses_levels);
    
    if(current!=NULL && nxt!=NULL)
    { 
        int sz= solver->level; 
        binary_map = (int *) calloc (sz, sizeof(int)); 
               
        CLEAR_STACK(solver->decision_conf_data.lbp);       

        compute_first_common(solver, current,nxt);        
        if(SIZE_STACK(solver->decision_conf_data.lbp))
        {          
            while(SIZE_STACK(solver->decision_conf_data.learned_clauses_levels))
            {
                current = POP_STACK(solver->decision_conf_data.learned_clauses_levels);                
                compute_rest_common(solver, current);
                if(SIZE_STACK(solver->decision_conf_data.lbp)==0)
                    break;
            }
        }
        free(binary_map);  
        CLEAR_STACK(solver->decision_conf_data.learned_clauses_levels);  
    }
}

void
kissat_last_decision_aggregate (kissat * solver)
{
    track_conflict_decision(solver);
    if(solver->decision_conf_data.learnts_last_decision>1)
    {
        compute_lbp(solver);
        tag_lbp_vars(solver);
        //printf("\n==>%d",solver->decision_conf_data.min_glue);        
    }
    solver->decision_conf_data.confs_last_decision = 0;
    solver->decision_conf_data.learnts_last_decision = 0;
    solver->decision_conf_data.min_glue = INT_MAX;
}