/****************************************************************************************
Copyright (c) 2020-2021, Saeed Nejati, Vijay Ganesh

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/
#ifndef _INITIALIZER_H_
#define _INITIALIZER_H_

#include <vector>
#include <map>
#include "core/SolverTypes.h"
//#include "core/Solver.h"

using namespace std;

struct BetaDist {
    double a, b;
};

#define BayesianWeight(v) (max(parameters[v].a, parameters[v].b) / (parameters[v].a + parameters[v].b))


/** Polarity and Activity Initializer for MiniSat based SAT solvers
 * Current methods:
 * 1. Default (All-zero / All-False)
 * 2. Bayesian Moment Matching
 * 3. Jeroslow-Wang
 * 4. Random
 */

namespace Minisat {

class Solver;

class SearchInitializer {
    public:

        enum InitMethod {
            DEFAULT,
            BMM,
            JW,
            RANDOM,
        };

        SearchInitializer();
        SearchInitializer(Solver* s, int polarity, int activity, int initEpoch, int updateEpoch);
        virtual ~SearchInitializer();

        void update(vector<Lit>& clause);

    private:
        void bayesian();
        template<typename T>
        void bayesian_update(T& c);
        void init_bayesian();

        void jeroslow_wang();
        
        Solver* solver;

        InitMethod polarity_init_method, activity_init_method;

        // BMM
        vector<BetaDist> parameters;
        int init_epochs;
        int update_epochs;
};

}
#endif
